@@ -89,15 +89,15 @@ sub new {
   my ($class, %options) = @_;
   my $self = bless {    # table => {},
     value   => {}, meaning  => {}, stash  => {}, stash_active => {},
-    catcode => {}, mathcode => {}, sfcode => {}, lccode => {}, uccode => {}, delcode => {},
+    catcode => {}, mathcode => {}, sfcode => {}, lccode       => {}, uccode => {}, delcode => {},
     undo    => [{ _FRAME_LOCK_ => 1 }], prefixes => {}, status => {},
-    stomach => $options{stomach}, model => $options{model} }, $class;
+    stomach => $options{stomach},       model    => $options{model} }, $class;
   # Note that "100" is hardwired into TeX, The Program!!!
   $$self{value}{MAX_ERRORS} = [100];
   # Standard TeX units, in scaled points
   $$self{value}{UNITS} = [{
-      pt => 65536, pc => 12 * 65536, in => 72.27 * 65536, bp => 72.27 * 65536 / 72,
-      cm => 72.27 * 65536 / 2.54, mm => 72.27 * 65536 / 2.54 / 10, dd => 1238 * 65536 / 1157,
+      pt => 65536,                    pc => 12 * 65536, in => 72.27 * 65536, bp => 72.27 * 65536 / 72,
+      cm => 72.27 * 65536 / 2.54,     mm => 72.27 * 65536 / 2.54 / 10, dd => 1238 * 65536 / 1157,
       cc => 12 * 1238 * 65536 / 1157, sp => 1,
       px => 72.27 * 65536 / 72,    # Assume px=bp ?
   }];
@@ -470,6 +470,8 @@ sub lookupDigestableDefinition {
       && ($entry      = $$self{meaning}{$lookupname})) {
       $defn = $$entry[0]; }
     return $defn; }
+  elsif ($CATCODE_ACTIVE_OR_CS[$cc]) {
+    return; }
   return $token; }
 
 # And a shorthand for installing definitions
@@ -706,9 +708,9 @@ sub convertUnit {
   $unit = lc($unit);
   # Put here since it could concievably evolve to depend on the current font.
   # Eventually try to track font size?
-  if    ($unit eq 'em') { return 10.0 * 65536; }
-  elsif ($unit eq 'ex') { return 4.3 * 65536; }
-  elsif ($unit eq 'mu') { return 10.0 * 65536 / 18; }
+  if    ($unit eq 'em') { return $self->lookupValue('font')->getEMWidth; }
+  elsif ($unit eq 'ex') { return $self->lookupValue('font')->getEXHeight; }
+  elsif ($unit eq 'mu') { return $self->lookupValue('font')->getMUWidth; }
   else {
     my $units = $self->lookupValue('UNITS');
     my $sp    = $$units{$unit};
@@ -745,7 +747,7 @@ sub getStatusMessage {
   my @undef        = ($$status{undefined} ? keys %{ $$status{undefined} } : ());
   my $undef_status = @undef && colorizeString(scalar(@undef) . " undefined macro" . (@undef > 1 ? 's' : '')
       . "[" . join(', ', @undef) . "]", 'details');
-  my @miss = ($$status{missing} ? keys %{ $$status{missing} } : ());
+  my @miss           = ($$status{missing} ? keys %{ $$status{missing} } : ());
   my $missing_status = @miss && colorizeString(scalar(@miss) . " missing file" . (@miss > 1 ? 's' : '')
       . "[" . join(', ', @miss) . "]", 'details');
 
