@@ -20,7 +20,7 @@ use IO::String;
 use Archive::Zip qw(:CONSTANTS :ERROR_CODES);
 
 use base qw(Exporter);
-our @EXPORT = qw(&unpack_source &pack_collection);
+our @EXPORT                       = qw(&unpack_source &pack_collection);
 our $archive_file_exclusion_regex = qr/(?:^\.)|(?:\.(?:zip|gz|epub|tex|bib|mobi|cache)$)|(?:~$)/;
 
 sub unpack_source {
@@ -39,21 +39,43 @@ sub unpack_source {
   # Extract the Perl zip datastructure to the temporary directory
   foreach my $member ($zip_handle->memberNames()) {
     $zip_handle->extractMember($member, catfile($sandbox_directory, $member)); }
-  # Set $source to point to the main TeX file in that directory
-  my @TeX_file_members = map { $_->fileName() } $zip_handle->membersMatching('\.tex$');
-  if (!@TeX_file_members) { # No .tex file? Try files with no, or unusually long, extensions
-    @TeX_file_members = grep {!/\./ || /\.[^.]{4,}$/} map { $_->fileName() } $zip_handle->members();
+
+  # I. Detect and return the main TeX file in that directory (or .txt, for old arXiv bundles)
+
+  # I.1. arXiv has a special metadata file identifying the primary source, and ignoring assets
+  if (my $readme_member = $zip_handle->memberNamed('00README.XXX')) {
+    my $readme_file = catfile($sandbox_directory, $readme_member->fileName());
+    open(my $README_FH, '<', $readme_file) or
+      (print STDERR "failed to open '$readme_file' for use as ZIP readme: $!. Continuing.\n");
+    local $/ = "\n";
+    my $toplevelfile;
+    while (<$README_FH>) {
+      chomp($_);
+      my ($name, $directive) = split(/\s+/, $_);
+      if ($directive eq 'toplevelfile') {
+        # shortcut guessing the top file, the user has provided it explicitly.
+        $toplevelfile = catfile($sandbox_directory, $name);
+      } elsif ($directive eq 'ignore') {
+        my $ignored_filepath = catfile($sandbox_directory, $name);
+        unlink($ignored_filepath) if -e $ignored_filepath; } }
+    return $toplevelfile if $toplevelfile; }
+
+  # I.2. Without an explicit directive,
+  #      heuristically determine the input (borrowed from arXiv::FileGuess)
+  my @TeX_file_members = map { $_->fileName() } $zip_handle->membersMatching('\.[tT](:?[eE][xX]|[xX][tT])$');
+  if (!@TeX_file_members) {    # No .tex file? Try files with no, or unusually long, extensions
+    @TeX_file_members = grep { !/\./ || /\.[^.]{4,}$/ } map { $_->fileName() } $zip_handle->members();
   }
 
-  # Heuristically determine the input (borrowed from arXiv::FileGuess)
-  my %Main_TeX_likelihood;
+  my (%Main_TeX_likelihood, %Main_TeX_level);
   my @vetoed = ();
   foreach my $tex_file (@TeX_file_members) {
     # Read in the content
     $tex_file = catfile($sandbox_directory, $tex_file);
     # Open file and read first few bytes to do magic sequence identification
     # note that file will be auto-closed when $FILE_TO_GUESS goes out of scope
-    open(my $FILE_TO_GUESS, '<', $tex_file) ||
+    next unless -e $tex_file;    # skip deleted "ignored" files.
+    open(my $FILE_TO_GUESS, '<', $tex_file) or
       (print STDERR "failed to open '$tex_file' to guess its format: $!. Continuing.\n");
     local $/ = "\n";
     my ($maybe_tex, $maybe_tex_priority, $maybe_tex_priority2);
@@ -81,9 +103,10 @@ sub unpack_source {
         # (it could in very elaborate multi-target setups, but we DON'T support those)
         # so veto it.
         my $vetoed_file = $1;
-        if ($vetoed_file eq 'amstex') { # TeX Priority
+        if ($vetoed_file eq 'amstex') {    # TeX Priority
           $Main_TeX_likelihood{$tex_file} = 2; last TEX_FILE_TRAVERSAL; }
         if ($vetoed_file !~ /\./) {
+          $vetoed_file =~ s/\s+$//;        # drop trailing spaces if any;
           $vetoed_file .= '.tex';
         }
         my $base = $tex_file;
@@ -95,21 +118,21 @@ sub unpack_source {
       if (/\\(?:end|bye)(?:\s|$)/) {
         $maybe_tex_priority2 = 1; }
       if (/\\input *(?:harv|lanl)mac/ || /\\input\s+phyzzx/) {
-        $Main_TeX_likelihood{$tex_file} = 1; last TEX_FILE_TRAVERSAL; }        # Mac TeX
+        $Main_TeX_likelihood{$tex_file} = 1; last TEX_FILE_TRAVERSAL; }    # Mac TeX
       if (/beginchar\(/) {
-        $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }        # MetaFont
+        $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }    # MetaFont
       if (/(?:^|\r)\@(?:book|article|inbook|unpublished)\{/i) {
-        $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }        # BibTeX
+        $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }    # BibTeX
       if (/^begin \d{1,4}\s+[^\s]+\r?$/) {
         if ($maybe_tex_priority) {
-          $Main_TeX_likelihood{$tex_file} = 2; last TEX_FILE_TRAVERSAL; }      # TeX Priority
+          $Main_TeX_likelihood{$tex_file} = 2; last TEX_FILE_TRAVERSAL; }    # TeX Priority
         if ($maybe_tex) {
-          $Main_TeX_likelihood{$tex_file} = 1; last TEX_FILE_TRAVERSAL; }      # TeX
-        $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }        # UUEncoded or PC
+          $Main_TeX_likelihood{$tex_file} = 1; last TEX_FILE_TRAVERSAL; }    # TeX
+        $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }      # UUEncoded or PC
       if (m/paper deliberately replaced by what little/) {
         $Main_TeX_likelihood{$tex_file} = 0; last TEX_FILE_TRAVERSAL; }
     }
-    close $FILE_TO_GUESS || warn "couldn't close file: $!";
+    close $FILE_TO_GUESS or (print STDERR "couldn't close file: $!");
     if (!defined $Main_TeX_likelihood{$tex_file}) {
       if ($maybe_tex_priority) {
         $Main_TeX_likelihood{$tex_file} = 2; }
@@ -122,22 +145,67 @@ sub unpack_source {
     }
   }
   # Veto files that were e.g. arguments of \input macros
-  for my $filename(@vetoed) {
-    delete $Main_TeX_likelihood{$filename};
-  }
-  # The highest likelihood (>0) file gets to be the main source.
-  my @files_by_likelihood = sort { $Main_TeX_likelihood{$b} <=> $Main_TeX_likelihood{$a} } grep { $Main_TeX_likelihood{$_} > 0 } keys %Main_TeX_likelihood;
+  for my $filename (@vetoed) {
+    delete $Main_TeX_likelihood{$filename}; }
+  # Examine only the max-scoring candidates (if any)
+  my @files_by_likelihood = sort { $Main_TeX_likelihood{$b} <=> $Main_TeX_likelihood{$a} }
+    grep { $Main_TeX_likelihood{$_} > 0 } keys %Main_TeX_likelihood;
   if (@files_by_likelihood) {
-   # If we have a tie for max score, grab the alphanumerically first file (to ensure deterministic runs)
     my $max_likelihood = $Main_TeX_likelihood{ $files_by_likelihood[0] };
-    @files_by_likelihood = sort { $a cmp $b } grep { $Main_TeX_likelihood{$_} == $max_likelihood } @files_by_likelihood;
+    @files_by_likelihood = grep { $Main_TeX_likelihood{$_} == $max_likelihood } @files_by_likelihood;
+# Special heuristic 1: If we have multiple "best" candidates, prefer those closer to the archive root '/'
+    if (scalar(@files_by_likelihood) > 1) {
+      my $min_count = 100;
+      foreach my $file (@files_by_likelihood) {
+        my $count = $file =~ tr/\///;
+        $Main_TeX_level{$file} = $count;
+        $min_count = $count if $min_count > $count; }
+      @files_by_likelihood = grep { $Main_TeX_level{$_} == $min_count } @files_by_likelihood; }
+    # Special heuristic 2: prefer candidates with PDF-like \includegraphics
+    if (scalar(@files_by_likelihood) > 1) {
+      my @pdf_includes = heuristic_check_for_pdftex(@files_by_likelihood);
+      @files_by_likelihood = @pdf_includes if @pdf_includes; }
+    # Special heuristic 3: prefer "best" candidates with a .bbl file
+    if (scalar(@files_by_likelihood) > 1) {
+      my @with_bbl = grep { my $base = $_; $base =~ s/\.tex$//; -e "$base.bbl"; } @files_by_likelihood;
+      @files_by_likelihood = @with_bbl if @with_bbl; }
+# Special heuristic 4 ?!: Sometimes in arXiv the decision is made in an unclear manner
+# (example: see 2112.08935 v1, which has equally good main.tex and bare_adv.tex)
+# so, for now, err on the side of preferring one of the extremely common names, when they are available at highest score.
+    if (scalar(@files_by_likelihood) > 1) {
+      my @common_name = grep { /(^|\W)(?:main|ms|paper)\.tex$/ } @files_by_likelihood;
+      @files_by_likelihood = @common_name if @common_name; }
+    # Final heuristic: tie-breaker is lexicographical order
+    if (scalar(@files_by_likelihood) > 1) {
+      @files_by_likelihood = sort { $a cmp $b } @files_by_likelihood; }
+    # set the winner for main source
     $main_source = shift @files_by_likelihood; }
 
   # If failed, clean up sandbox directory.
   rmtree($sandbox_directory) unless $main_source;
   # Return the main source from the unpacked files in the sandbox directory (or undef if failed)
-  return $main_source;
-}
+  return $main_source; }
+
+sub heuristic_check_for_pdftex {
+  my @filenames    = @_;
+  my @pdf_includes = ();
+  for my $tex_file (@filenames) {
+    my $is_open = open(my $TEX_FH, '<', $tex_file);
+    if (!$is_open) {
+      print STDERR "failed to open '$tex_file' to guess its format: $!. Continuing.\n";
+      next; }
+    local $/ = "\n";
+    my $pdfoutput_checks = 5;
+    while (<$TEX_FH>) {
+      # Check from arXiv::FileGuess, line 317
+      # https://metacpan.org/release/SIMEON/arXiv-1.01/source/lib/arXiv/FileGuess.pm#L317
+      if (/^[^%]*\\includegraphics[^%]*\.(?:pdf|png|gif|jpg)\s?\}/i ||
+        ($pdfoutput_checks >= 0 && /^[^%]*\\pdfoutput(?:\s+)?=(?:\s+)?1/)) {
+        push(@pdf_includes, $tex_file);
+        last; }
+      $pdfoutput_checks-- if $pdfoutput_checks; }
+    close $TEX_FH or (print STDERR "couldn't close file: $!"); }
+  return @pdf_includes; }
 
 # Options:
 #   whatsout: determine what shape and size we want to pack into
@@ -202,15 +270,17 @@ sub get_archive {
       or (print STDERR "Fatal:I/O:$pathname File $pathname is not readable.");
     my $file_contents = <$FH>;
     close($FH);
-    # Only compress the textual files
-    my $compression_level = ($file =~ /css|js|xml|html$/) ?
-      COMPRESSION_DEFLATED
-      : COMPRESSION_STORED;
-    $archive->addString($file_contents, $file,)->desiredCompressionMethod($compression_level); }
+    # Compress all files except mimetype
+    my $compression = ($file eq 'mimetype' ? COMPRESSION_STORED : COMPRESSION_DEFLATED);
+    $archive->addString($file_contents, $file,)->desiredCompressionMethod($compression); }
 
   foreach my $subdir (sort @subdirs) {
     my $current_dir = File::Spec->catdir($directory, $subdir);
-    $archive->addTree($current_dir, $subdir, sub { !/$archive_file_exclusion_regex/ }, COMPRESSION_STORED); }
+    $archive->addTree($current_dir, $subdir, sub { !/$archive_file_exclusion_regex/ }, COMPRESSION_DEFLATED); }
+
+  if (defined $ENV{SOURCE_DATE_EPOCH}) {
+    for my $member ($archive->members()) {
+      $member->setLastModFileDateTimeFromUnix($ENV{SOURCE_DATE_EPOCH}); } }
 
   my $payload;
   if ($whatsout =~ /^archive(::zip)?$/) {
@@ -226,9 +296,9 @@ sub get_math {
   return unless defined $doc;
   my @mnodes     = $doc->findnodes($math_xpath);
   my $math_count = scalar(@mnodes);
-  if (!$math_count) { # If no real math nodes, look for math image nodes
-    my $math_img_xpath = '//*[local-name()="img" and @class="ltx_Math"]';
-    @mnodes         = $doc->findnodes($math_img_xpath);
+  if (!$math_count) {    # If no real math nodes, look for math image nodes
+    my $math_img_xpath = '//*[local-name()="img" and contains(@class,"ltx_Math")]';
+    @mnodes     = $doc->findnodes($math_img_xpath);
     $math_count = scalar(@mnodes);
   }
   if (!$math_count) {
@@ -238,7 +308,7 @@ sub get_math {
     my $math_found = 0;
     while ($math_found != $math_count) {
       $math_found = $math->findnodes('.' . $math_xpath)->size;
-      $math_found++ if ($math->localname =~ /^math$/i);
+      $math_found++             if ($math->localname =~ /^math$/i);
       $math = $math->parentNode if ($math_found != $math_count);
     }
     $math = $math->parentNode while ($math->nodeName =~ '^t[rd]$'); }
